//@line 47 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\ThumbnailProtocol.js"

"use strict";

const Cu = Components.utils;
const Cc = Components.classes;
const Cr = Components.results;
const Ci = Components.interfaces;

// define the scheme to use - "moz-thumb://..."
const SCHEME = "moz-thumb";

// define the image format of our thumbnails
const CONTENT_TYPE = "image/png";

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

XPCOMUtils.defineLazyServiceGetter(this, "gAppShellService",
                                   "@mozilla.org/appshell/appShellService;1",
                                   "nsIAppShellService");

XPCOMUtils.defineLazyServiceGetter(this, "gCacheService",
                                   "@mozilla.org/network/cache-service;1",
                                   "nsICacheService");

XPCOMUtils.defineLazyGetter(this, "gNetUtil", function () {
  var obj = {};
  Cu.import("resource://gre/modules/NetUtil.jsm", obj);
  return obj.NetUtil;
});

//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\helper.js"

// ----------
// Function: debug
// Logs a given debug message to the console.
//
// Parameters:
//   aMsg - the log message
function debug(aMsg) {
  aMsg = ("Thumbnails: " + aMsg).replace(/\S{80}/g, "$&\n");
  Services.console.logStringMessage(aMsg);
}

// ----------
// Function: trace
// Logs a given debug message to the console and prints the current stack trace.
//
// Parameters:
//   aMsg - the log message
function trace(aMsg) {
  // cut off the first line of the stack trace, because that's just this function.
  let stack = Error().stack.split("\n").slice(1);

  debug("trace: " + aMsg + "\n" + stack.join("\n"));
}

// ----------
// Function: parseUri
// Parses the given URI and returns its parameters.
//
// Parameters:
//   aUri - the URI to parse
function parseUri(aUri) {
  let params = {};
  let [host, query] = aUri.spec.split("?");

  query.split("&").forEach(function (param) {
    let [key, value] = param.split("=").map(decodeURIComponent);
    params[key.toLowerCase()] = value;
  });

  params.algorithm = host.replace(/^.*\/+/, "").replace(/\/+.*$/, "");

  return params;
}

// ----------
// Function: getCacheKey
// Returns the cache key to use for the given thumbnail URI. This puts params
// in a fixed order and removes params that must not be included in the key.
//
// Parameters:
//   aUri - the URI to get the cache key for
function getCacheKey(aUri) {
  let keys = ["algorithm", "url", "width", "height"];
  let params = parseUri(aUri);

  let pairs = keys.map(function (key) {
    let value = encodeURIComponent(params[key] || "");
    return encodeURIComponent(key) + "=" + value;
  });

  return pairs.join("&");
}
//@line 40 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\protocol.js"

// a white 1x1 image that is delivered as a thumbnail for about:blank
const ABOUT_BLANK_URI = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAA" +
                        "BCAYAAAAfFcSJAAAADUlEQVQImWP4////fwAJ+wP9CNHoHgAAAABJ" +
                        "RU5ErkJggg==";

// specifies the flags specific to our protocol implementation
const PROTOCOL_FLAGS = Ci.nsIProtocolHandler.URI_DANGEROUS_TO_LOAD |
                       Ci.nsIProtocolHandler.URI_NORELATIVE |
                       Ci.nsIProtocolHandler.URI_NOAUTH;

// ##########
// Class: Protocol
// Implements the thumbnail protocol handler responsible for moz-thumb:// URIs.
function Protocol() {
}

Protocol.prototype = {
  get scheme() SCHEME,
  get defaultPort() -1,
  get protocolFlags() PROTOCOL_FLAGS,

  // ----------
  // Function: newURI
  // Creates a new URI object that is suitable for loading by this protocol.
  //
  // Parameters:
  //   aSpec - URI string in UTF8 encoding
  //   aOriginCharset - charset of the document from which the URI originated
  newURI: function Proto_newURI(aSpec, aOriginCharset) {
    let uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
    uri.spec = aSpec;
    return uri;
  },

  // ----------
  // Function: newChannel
  // Constructs a new channel from the given URI for this protocol handler.
  //
  // Parameters:
  //   aUri - the URI for which to construct a channel
  newChannel: function Proto_newChannel(aUri) {
    let {url} = parseUri(aUri);

    // don't try to load thumbnails for about:blank pages, just deliver a
    // white 1x1 image. so we make sure that tests can use about:blank
    // and don't have to wait until all the images have been rendered.
    if (0 == url.indexOf("about:blank"))
      return Services.io.newChannel(ABOUT_BLANK_URI, "UTF8", null);

    // prevent recursion if a thumbnail for a thumbnail is requested
    if (0 == url.indexOf(SCHEME + "://"))
      aUri = Services.io.newURI(url, "UTF8", null);

    // return a redirect channel that will asynchronously decide which channel
    // to redirect to.
    return new Channel(aUri);
  },

  // ----------
  // Function: allowPort
  // Returns always false because we'll never want to allow blacklisted ports.
  //
  // Parameters:
  //   aPort - the blacklisted port
  //   aScheme - the scheme the port is blacklisted for
  allowPort: function Proto_allowPort(aPort, aScheme) {
    return false;
  },

  classID: Components.ID("{5a4ae9b5-f475-48ae-9dce-0b4c1d347884}"),
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler])
};

let NSGetFactory = XPCOMUtils.generateNSGetFactory([Protocol]);
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\channel.js"

function Channel(aUri) {
  this._uri = aUri;

  // nsIChannel
  this.originalURI = aUri;

  // nsIHttpChannel
  this._requestHeaders = {};
  this._responseHeaders = {"content-type": CONTENT_TYPE};
}

Channel.prototype = {
  _wasOpened: false,
  _wasStarted: false,
  _isCanceled: false,

  // ----------
  // Function: asyncOpen
  // Opens this channel asynchronously.
  //
  // Parameters:
  //   aListener - the listener that receives the channel data when available
  //   aContext - a custom context passed to the listener's methods
  asyncOpen: function CH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    let {refresh} = parseUri(this.URI);
    this._cacheKey = getCacheKey(this.URI);

    // the url specifies that we should refresh the thumbnail. so let's
    // create a new thumbnail for the given url and store it in the cache
    // regardless of whether it's already cached.
    if (refresh) {
      GenerateChannel.extend(this);
      return;
    }

    let self = this;

    // query the cache and retrieve our entry
    Cache.getReadEntry(this._cacheKey, function (aEntry) {
      let inputStream = aEntry && aEntry.openInputStream(0);

      if (aEntry && inputStream.available())
        // entry found, deliver from cache
        CacheChannel.extend(self, aEntry);
      else
        // entry not found, create a thumbnail and deliver it when ready
        GenerateChannel.extend(self);
    });
  },

  // ----------
  // Function: _startRequest
  // Calls onStartRequest on the channel listener.
  _startRequest: function BCH__startRequest() {
    try {
      this._listener.onStartRequest(this, this._context);
    } catch (e) {
      // the listener might throw if the request has been canceled
      this.cancel(Cr.NS_BINDING_ABORTED);
    }

    return !this.canceled;
  },

  // ----------
  // Function: _serveData
  // Calls onDataAvailable on the channel listener and passes the data stream.
  //
  // Parameters:
  //   aUri - the URI belonging to this channel
  _serveData: function BCH__serveData(aDataStream) {
    try {
      let available = aDataStream.available();
      this._listener.onDataAvailable(this, this._context, aDataStream, 0, available);
    } catch (e) {
      // the listener might throw if the request has been canceled
      this.cancel(Cr.NS_BINDING_ABORTED);
    }

    return !this.canceled;
  },

  // ----------
  // Function: _stopRequest
  // Calls onStopRequest on the channel listener.
  _stopRequest: function BCH__stopRequest() {
    try {
      this._listener.onStopRequest(this, this._context, this.status);
    } catch (e) {
      // this might throw but is generally ignored
    }

    // the request has finished, clean up after ourselves
    this._cleanup();
  },

  // ----------
  // Function: _addToLoadGroup
  // Adds this request to the load group, if any.
  _addToLoadGroup: function BCH__addToLoadGroup() {
    if (this.loadGroup)
      this.loadGroup.addRequest(this, this._context);
  },

  // ----------
  // Function: _removeFromLoadGroup
  // Removes this request from the load group, if any.
  _removeFromLoadGroup: function BCH__removeFromLoadGroup() {
    if (!this.loadGroup)
      return;

    try {
      this.loadGroup.removeRequest(this, this._context, this.status);
    } catch (e) {
      // this might throw and is ignored
    }
  },

  // ----------
  // Function: _cleanup
  // Cleans up the channel when the request has finished.
  _cleanup: function BCH__cleanup() {
    this._removeFromLoadGroup();
    this.loadGroup = null;

    this._isPending = false;

    if (this._entry)
      this._entry.close();

    delete this._listener;
    delete this._context;
    delete this._entry;
  },

  // nsIChannel
  contentType: CONTENT_TYPE,
  contentLength: -1,
  owner: null,
  contentCharset: null,
  notificationCallbacks: null,

  get URI() this._uri,
  get securityInfo() null,

  // ----------
  // Function: open
  // Opens this channel synchronously.
  open: function BCH_open() {
    // synchronous data delivery is not implemented
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  // nsIHttpChannel
  redirectionLimit: 10,
  requestMethod: "GET",
  allowPipelining: true,
  referrer: null,

  get requestSucceeded() true,
  get responseStatus() 200,
  get responseStatusText() "OK",

  // ----------
  // Function: isNoCacheResponse
  // Returns true if the server sent the equivalent of a
  // "Cache-control: no-cache" response header. Always false.
  isNoCacheResponse: function () false,

  // ----------
  // Function: isNoStoreResponse
  // Returns true if the server sent the equivalent of a
  // "Cache-control: no-cache" response header. Always false.
  isNoStoreResponse: function () false,

  // ----------
  // Function: getRequestHeader
  // Returns the value of a particular request header.
  //
  // Parameters:
  //   aHeader - the case-insensitive name of the request header to query
  getRequestHeader: function BCH_getRequestHeader(aHeader) {
    aHeader = aHeader.toLowerCase();
    let headers = this._requestHeaders;

    if (aHeader in headers)
      return headers[aHeader];

    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  // ----------
  // Function: setRequestHeader
  // This method is called to set the value of a particular request header.
  //
  // Parameters:
  //   aHeader - the case-insensitive name of the request header to query
  //   aValue - the request header value to set
  setRequestHeader: function BCH_setRequestHeader(aHeader, aValue, aMerge) {
    if (aMerge)
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    this._requestHeaders[aHeader.toLowerCase()] = aValue;
  },

  // ----------
  // Function: visitRequestHeaders
  // Call this method to visit all request headers.
  //
  // Parameters:
  //   aVisitor - the header visitor
  visitRequestHeaders: function BCH_visitRequestHeaders(aVisitor) {
    this._visitHeaders(this._requestHeaders, aVisitor);
  },

  // ----------
  // Function: getResponseHeader
  // Returns the value of a particular response header.
  //
  // Parameters:
  //   aHeader - the case-insensitive name of the response header to query
  getResponseHeader: function BCH_getResponseHeader(aHeader) {
    aHeader = aHeader.toLowerCase();
    let headers = this._responseHeaders;

    if (aHeader in headers)
      return headers[aHeader];

    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  // ----------
  // Function: setResponseHeader
  // This method is called to set the value of a particular response header.
  //
  // Parameters:
  //   aHeader - the case-insensitive name of the response header to query
  //   aValue - the response header value to set
  setResponseHeader: function BCH_setResponseHeader(aHeader, aValue, aMerge) {
    if (aMerge)
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    this._responseHeaders[aHeader.toLowerCase()] = aValue;
  },

  // ----------
  // Function: visitResponseHeaders
  // Call this method to visit all response headers.
  //
  // Parameters:
  //   aVisitor - the header visitor
  visitResponseHeaders: function BCH_visitResponseHeaders(aVisitor) {
    this._visitHeaders(this._responseHeaders, aVisitor);
  },

  // ----------
  // Function: _visitHeaders
  // Generic function to visit an array of headers.
  //
  // Parameters:
  //   aHeaders - the headers to visit
  //   aVisitor - the header visitor
  _visitHeaders: function BCH__visitHeaders(aHeaders, aVisitor) {
    for (let name in aHeaders) {
      let value = aHeaders[name];

      try {
        aVisitor.visitHeader(name, value);
      } catch (e) {
        // the visitor throws to stop the iteration
        return;
      }
    }
  },

  // nsIRequest
  _status: Cr.NS_OK,
  _isPending: false,

  loadFlags: Ci.nsIRequest.LOAD_NORMAL,
  loadGroup: null,

  get name() this._uri.spec,
  get status() this._status,

  isPending: function () this._isPending,
  resume: function () {},
  suspend: function () {},

  // ----------
  // Function: cancel
  // Cancels this request.
  //
  // Parameters:
  //   aStatus - the reason for cancelling
  cancel: function BCH_cancel(aStatus) {
    if (this.canceled)
      return;

    this._isCanceled = true;
    this._status = aStatus;

    this._cleanup();
  },

  // nsIHttpChannelInternal
  documentURI: null,

  get canceled() this._isCanceled,

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIChannel,
                                         Ci.nsIHttpChannel,
                                         Ci.nsIHttpChannelInternal,
                                         Ci.nsIRequest])
};
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\cacheChannel.js"

// ##########
// Class: CacheChannel
// A http channel implementation that delivers a thumbnail specified by the
// given URI from the image cache.
let CacheChannel = {
  // ----------
  // Function: extend
  // Extends the given channel with our custom functionality.
  //
  // Parameters:
  //   aChannel - the channel to extend
  //   aEntry - the cache entry we read from
  extend: function CC_extend(aChannel, aEntry) {
    for (let name in this)
      if (name != "extend")
        aChannel[name] = this[name];

    try {
      // set the content-length header stored in the entry's meta data
      let contentLength = aEntry.getMetaDataElement("content-length");
      aChannel.setResponseHeader("content-length", contentLength);
    } catch (e) {
      // throws if the meta data element doesn't exist, ignore
    }

    aChannel._entry = aEntry;
    aChannel._serve();
  },

  // ----------
  // Function: _serve
  // Serves data from the cache entry to the channel listener.
  _serve: function CC__serve() {
    let self = this;

    this._readEntry(function (aDataStream) {
      self._dataStream = aDataStream;

      if (!self._startRequest())
        return;

      self._addToLoadGroup();

      if (self._serveData(aDataStream))
        self._stopRequest();
    });
  },

  // ----------
  // Function: _readEntry
  // Reads a data stream from the cache entry.
  //
  // Parameters:
  //   aCallback - the callback the data is passed to
  _readEntry: function CC__readEntry(aCallback) {
    let self = this;
    let inputStream = this._entry.openInputStream(0);

    gNetUtil.asyncFetch(inputStream, function (aDataStream, aStatus, aRequest) {
      inputStream.close();

      // we might have been canceled while waiting
      if (self.canceled)
        return;

      // check if we have a valid data stream
      if (!Components.isSuccessCode(aStatus) || !aDataStream.available()) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      aCallback(aDataStream);
    });
  }
};
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\generateChannel.js"

// ##########
// Class: GenerateChannel
// A http channel implementation that creates a thumbnail for the requested URI
// delivers it and stores it in the cache.
let GenerateChannel = {
  // ----------
  // Function: extend
  // Extends the given channel with our custom functionality.
  //
  // Parameters:
  //   aChannel - the channel to extend
  extend: function GC_extend(aChannel) {
    for (let name in this)
      if (name != "extend")
        aChannel[name] = this[name];

    aChannel._serve();
  },

  // ----------
  // Function: _serve
  // Serves the thumbnails's data and writes it to the entry's output stream.
  _serve: function GC__serve() {
    let self = this;

    Cache.getWriteEntry(this._cacheKey, function (aEntry) {
      // store the entry so that _cleanup can close it
      self._entry = aEntry;

      // we might've been canceled while waiting for the entry
      if (self.canceled)
        return;

      // for some reason we don't have a valid entry
      if (!aEntry) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      // the listener might cancel when starting the request
      if (!self._startRequest())
        return;

      self._addToLoadGroup();

      // create our thumbnail and wait for its data
      self._createThumbnail(function (aDataStream, aStatusCode) {
        // store thumbnails in the cache only on success
        if (self._isSuccessStatusCode(aStatusCode))
          self._prepareCacheEntry(aEntry, aDataStream);

        // serve the image data to the listener
        if (self._serveData(aDataStream))
          self._stopRequest();
      });
    });
  },

  // ----------
  // Function: _isSuccessStatusCode
  // Returns whether the given status code denotes a success.
  //
  // Parameters:
  //   aStatusCode - the http status code to check
  _isSuccessStatusCode: function GC__isSuccessStatusCode(aStatusCode) {
    // return true for all 2xx status codes
    return Math.floor(aStatusCode / 100) == 2;
  },

  // ----------
  // Function: _createThumbnail
  // Creates a thumbnail and calls the given callback with a data stream and a
  // http status code.
  //
  // Parameters:
  //   aCallback - the callback to be called with the thumbnail's data stream
  //               and its http status code
  _createThumbnail: function GC__createThumbnail(aCallback) {
    let self = this;
    let {url, width, height} = parseUri(this._uri);

    // create a new thumbnail
    Thumbnailer.create(url, width, height, function (aDataStream, aStatusCode) {
      if (aDataStream && aDataStream.available())
        // valid thumbnail data
        aCallback(aDataStream, aStatusCode);
      else
        // failed to get the thumbnail data
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
    });
  },

  // ----------
  // Function: _prepareCacheEntry
  // Prepares a given cache entry so that we can write the thumbnail data
  // to its output stream.
  //
  // Parameters:
  //   aEntry - the entry we prepare to be written to
  //   aDataStream - the data stream that is written to the given entry
  _prepareCacheEntry: function GC__prepareCacheEntry(aEntry, aDataStream) {
    // create a tee listener to simultaneously deliver data to the
    // stream listener and to the cache entry's stream
    this._useWriteEntryTeeListener(aEntry);

    // store content-length header
    let available = "" + aDataStream.available();
    aEntry.setMetaDataElement("content-length", available);

    // store this entry for two days
    let unixTime = Math.floor(Date.now() / 1000);
    aEntry.setExpirationTime(unixTime + (86400 * 2));
  },

  // ----------
  // Function: _useWriteEntryTeeListener
  // Replaces the current listener with a TeeListener implementation that
  // simultaneously delivers the entry and writes it to the cache entry.
  //
  // Parameters:
  //   aEntry - the entry to whose data stream we write simultaneously
  _useWriteEntryTeeListener: function GC__useWriteEntryTeeListener(aEntry) {
    let tee = Cc["@mozilla.org/network/stream-listener-tee;1"]
              .createInstance(Ci.nsIStreamListenerTee);

    let outputStream = aEntry.openOutputStream(0);
    tee.init(this._listener, outputStream);

    this._listener = tee;
  }
};
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\cache.js"

// the cache client identifier to separate our cache content from others
const CACHE_CLIENT_ID = "moz-thumb";

// ##########
// Class: Cache
// Singleton providing access to the internal cache.
let Cache = {
  // lazy getter for the cache session
  get _cacheSession() {
    let session = gCacheService.createSession(CACHE_CLIENT_ID,
                                              Ci.nsICache.STORE_ON_DISK, true);

    delete this._cacheSession;
    return this._cacheSession = session;
  },

  // ----------
  // Function: getReadEntry
  // Calls the given <aCallback> with a cache entry opened for reading.
  //
  // Parameters:
  //   aKey - the key identifying the desired cache entry
  //   aCallback - the callback that is called when the cache entry is ready
  getReadEntry: function Cache_getReadEntry(aKey, aCallback) {
    let self = this;

    // try to open the desired cache entry
    this._openCacheEntry(aKey, Ci.nsICache.ACCESS_READ, function (aEntry) {
      if (!self._isValidReadEntry(aEntry))
        aEntry = null;

      aCallback(aEntry);
    });
  },

  // ----------
  // Function: getWriteEntry
  // Calls the given <aCallback> with a cache entry opened for writing.
  //
  // Parameters:
  //   aKey - the key identifying the desired cache entry
  //   aCallback - the callback that is called when the cache entry is ready
  getWriteEntry: function Cache_getWriteEntry(aKey, aCallback) {
    // try to open the desired cache entry
    this._openCacheEntry(aKey, Ci.nsICache.ACCESS_WRITE, aCallback);
  },

  // ----------
  // Function: _isValidReadEntry
  // Returns whether the given cache entry is valid and can be read from.
  //
  // Parameters:
  //   aEntry - the cache entry to verify
  _isValidReadEntry: function Cache__isValidReadEntry(aEntry) {
    if (!aEntry)
      return false;

    let inputStream = aEntry.openInputStream(0);

    // check if the entry actually contains data
    return inputStream && inputStream.available();
  },

  // ----------
  // Function: _openCacheEntry
  // Opens the cache entry identified by the given <aKey>.
  //
  // Parameters:
  //   aKey - the key identifying the desired cache entry
  //   aAccess - the desired access mode
  //   aCallback - the callback to be called when the cache entry was opened
  _openCacheEntry: function Cache__openCacheEntry(aKey, aAccess, aCallback) {
    function onCacheEntryAvailable(aEntry, aAccessGranted, aStatus) {
      let validAccess = (aAccess == aAccessGranted);
      let validStatus = Components.isSuccessCode(aStatus);

      // check if a valid entry was passed and if the access we requested
      // was actually granted
      if (aEntry && !(validAccess && validStatus)) {
        aEntry.close();
        aEntry = null;
      }

      aCallback(aEntry);
    }

    let listener = this._createCacheListener(onCacheEntryAvailable);
    this._cacheSession.asyncOpenCacheEntry(aKey, aAccess, listener);
  },

  // ----------
  // Function: _createCacheListener
  // Returns a cache listener implementing the nsICacheListener interface.
  //
  // Parameters:
  //   aCallback - the callback to be called when the cache entry is available
  _createCacheListener: function Cache__createCacheListener(aCallback) {
    return {
      onCacheEntryAvailable: aCallback,
      QueryInterface: XPCOMUtils.generateQI([Ci.nsICacheListener])
    };
  }
};
//@line 39 "e:\builds\moz2_slave\ux-w64-ntly\build\browser\components\thumbnails\thumbnailer.js"

// the XUL namespace used to create elements in XUL documents
const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

// as long as there's bug 565388, we need some lightweight chrome page
// to use as a host frame where we can create some XUL <iframes>
const XUL_FRAME_SRC = "chrome://global/content/mozilla.xhtml";

// the minimum width of the window we're generating thumbnails from
const THUMB_MIN_WIDTH = 1024;

// the color our thumbnails are using for transparent areas
const THUMB_BG_COLOR = "#fff";

// ##########
// Class: Thumbnailer
// Creates thumbnails for a given URL by loading it in an <iframe> in a hidden
// window, drawing the window contents to a canvas and finally providing the
// data stream to be read from.
let Thumbnailer = {
  // ----------
  // Function: create
  // Creates a thumbnail for a given url.
  //
  // Parameters:
  //   aUrl - the url to create thumbnail for
  //   aWidth - the thumbnail width
  //   aHeight - the thumbnail height
  //   aCallback - the callback to be called when the thumbnail has been created
  create: function TN_create(aUrl, aWidth, aHeight, aCallback) {
    // calculate our iframe size for a nice aspect ratio
    let frameWidth = Math.max(aWidth, THUMB_MIN_WIDTH);
    let frameHeight = frameWidth / (aWidth / aHeight);

    let self = this;

    // create a hidden frame containing the url to create a thumbnail for
    HiddenFrame.create(aUrl, frameWidth, frameHeight, function (aFrame, aStatusCode) {
      let window = aFrame.contentWindow;

      // create a canvas in the hidden frame
      let canvas = self._createCanvas(window, aWidth, aHeight);
      let ctx = canvas.getContext("2d");

      // scale canvas accordingly
      let scaler = aWidth / frameWidth;
      ctx.scale(scaler, scaler);

      // draw to the canvas as soon as the frame is initialized
      ctx.drawWindow(window, 0, 0, frameWidth, frameHeight,
                     THUMB_BG_COLOR, ctx.DRAWWINDOW_DO_NOT_FLUSH);

      // read the image data from the canvas and pass it to the callback
      self._readCanvasData(canvas, function (aDataStream) {
        aCallback(aDataStream, aStatusCode);
        HiddenFrame.destroy(aFrame);
      });
    });
  },

  // ----------
  // Function: _createCanvas
  // Creates a canvas in the given DOM window.
  //
  // Parameters:
  //   aWindow - the DOM window that contains the canvas
  //   aWidth - the canvas width
  //   aHeight - the canvas height
  _createCanvas: function TN__createCanvas(aWindow, aWidth, aHeight) {
    let doc = aWindow.document;
    let canvas = doc.createElement("canvas");
    doc.body.appendChild(canvas);

    canvas.setAttribute("moz-opaque", "");
    canvas.style.width = aWidth + "px";
    canvas.style.height = aHeight + "px";
    canvas.style.visibility = "hidden";
    canvas.width = aWidth;
    canvas.height = aHeight;

    return canvas;
  },

  // ----------
  // Function: _readCanvasData
  // Reads the image data from a given canvas and provides a data stream.
  //
  // Parameters:
  //   aCanvas - the canvas to read the image data from
  //   aCallback - the callback that the image data is passed to
  _readCanvasData: function TN__readCanvasData(aCanvas, aCallback) {
    let dataUri = aCanvas.toDataURL(CONTENT_TYPE, "");
    let uri = Services.io.newURI(dataUri, "UTF8", null);

    gNetUtil.asyncFetch(uri, aCallback);
  }
};

// ##########
// Class: HiddenFrame
// Creates a hidden <iframe> element that can be used to load and process
// web pages in the background.
let HiddenFrame = {
  // prevent a race condition and queue all _getHostFrame() requests if we're
  // creating a new one
  _hostFrameCallbacks: [],

  // ----------
  // Function: create
  // Creates a new hidden frame for a given uri with a given size.
  //
  // Parameters:
  //   aUri - the uri loaded in the hidden frame
  //   aWidth - the hidden frame's width
  //   aHeight - the hidden frame's height
  //   aCallback - the callback that the new hidden frame is passed to
  create: function HF_create(aUri, aWidth, aHeight, aCallback) {
    let self = this;

    this._createNewFrame(function (aFrame) {
      // set the iframe size
      let style = aFrame.style;
      style.width = aWidth + "px";
      style.height = aHeight + "px";

      // hide the scrollbars
      style.overflow = "hidden";

      // start loading the frame
      aFrame.setAttribute("src", aUri);

      // wait until the iframe's document has finished loading
      aFrame.addEventListener("pageshow", function onLoad(aEvent) {
        // we want the top document to be loaded and not some iframe contained
        // in our iframe
        if (aFrame.contentDocument != aEvent.target)
          return;

        aFrame.removeEventListener("pageshow", onLoad, false);

        // delay calling the callback to give the rendering engine time to
        // finish displaying the content in the iframe
        aFrame._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
        aFrame._timer.initWithCallback(function () {
          delete aFrame._timer;
          aCallback(aFrame, self._getFrameStatusCode(aFrame));
        }, 0, Ci.nsITimer.TYPE_ONE_SHOT);
      }, false);
    });
  },

  // ----------
  // Function: destroy
  // Destroys the given hidden frame.
  //
  // Parameters:
  //   aFrame - the hidden frame to destroy
  destroy: function HF_destroy(aFrame) {
    aFrame.parentNode.removeChild(aFrame);
  },

  // ----------
  // Function: _createNewFrame
  // Creates a new hidden frame.
  //
  // Parameters:
  //   aCallback - the callback that the new hidden frame is passed to
  _createNewFrame: function HF__createNewFrame(aCallback) {
    this._getHostFrame(function (aFrame) {
      let doc = aFrame.contentDocument;
      let iframe = doc.createElementNS(XUL_NS, "iframe");
      iframe.setAttribute("type", "content");
      doc.documentElement.appendChild(iframe);

      // set docShell properties to prevent the hidden iframe from misbehaving
      let docShell = iframe.docShell;
      docShell.allowAuth = false;
      docShell.allowWindowControl = false;

      aCallback(iframe);
    });
  },

  // ----------
  // Function: _getHostFrame
  // Gets the host frame that contains the hidden frames.
  //
  // Parameters:
  //   aCallback - the callback that the host frame is passed to
  _getHostFrame: function HF__getHostFrame(aCallback) {
    // return the cached value, if any
    if (this._hostFrame) {
      aCallback(this._hostFrame);
      return;
    }

    // if we're not the first item in the queue we don't need to do a thing
    // but wait for the host frame to be loaded
    if (this._hostFrameCallbacks.push(aCallback) > 1)
      return;

    // create the host frame
    let doc = gAppShellService.hiddenDOMWindow.document;
    let hostFrame = doc.createElement("iframe");
    doc.documentElement.appendChild(hostFrame);
    hostFrame.setAttribute("src", XUL_FRAME_SRC);

    let self = this;

    // wait until the host frame is loaded
    hostFrame.addEventListener("DOMContentLoaded", function onLoad() {
      hostFrame.removeEventListener("DOMContentLoaded", onLoad, false);

      // cache the new host frame
      self._hostFrame = hostFrame;

      // pass the newly created host frame to every callback in the queue
      self._hostFrameCallbacks.forEach(function (callback) callback(hostFrame));
      self._hostFrameCallbacks = null;
    }, false);
  },

  // ----------
  // Function: _getFrameStatusCode
  // Returns the http status code for a given iframe.
  //
  // Parameters:
  //   aFrame - the iframe to get the status code for
  _getFrameStatusCode: function HF__getFrameStatusCode(aFrame) {
    let channel = aFrame.docShell.currentDocumentChannel;

    try {
      // if the channel is a nsIHttpChannel get its http status
      let httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
      return httpChannel.responseStatus;
    } catch (e) {
      // not a http channel, we just assume a success status code
      return "200";
    }
  }
};
